<?php

namespace Drupal\commerce_packeta\Plugin\Commerce\CheckoutPane;

use Drupal\commerce\InlineFormManager;
use Drupal\commerce_checkout\Plugin\Commerce\CheckoutFlow\CheckoutFlowInterface;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_shipping\OrderShipmentSummaryInterface;
use Drupal\commerce_shipping\PackerManagerInterface;
use Drupal\commerce_shipping\Plugin\Commerce\CheckoutPane\ShippingInformation;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\profile\Entity\ProfileInterface;

/**
 * Provides the shipping information pane.
 *
 * Collects the shipping profile, then the information for each shipment.
 * Assumes that all shipments share the same shipping profile.
 *
 * @CommerceCheckoutPane(
 *   id = "packeta_capable_shipping_information",
 *   label = @Translation("Shipping information compatible with packeta"),
 *   display_label = @Translation("Shipping information"),
 *   wrapper_element = "fieldset",
 * )
 */
class PickupCapableShippingInformation extends ShippingInformation {

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, CheckoutFlowInterface $checkout_flow, EntityTypeManagerInterface $entity_type_manager, EntityTypeBundleInfo $entity_type_bundle_info, InlineFormManager $inline_form_manager, PackerManagerInterface $packer_manager, OrderShipmentSummaryInterface $order_shipment_summary) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $checkout_flow, $entity_type_manager, $entity_type_bundle_info, $inline_form_manager, $packer_manager, $order_shipment_summary);
    // Set requirements for shipping profile as false in order for our logic to work properly
    $this->configuration['require_shipping_profile'] = FALSE;
    $this->configuration['auto_recalculate'] = FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationSummary() {
    return '';
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {}

  /**
   * Sadly this functions code had to be copy pasted with only few replacements.
   *
   * {@inheritdoc}
   */
  public function buildPaneForm(array $pane_form, FormStateInterface $form_state, array &$complete_form) {
    $pane_form = parent::buildPaneForm($pane_form, $form_state, $complete_form);

    unset($pane_form['recalculate_shipping']);

    $store = $this->order->getStore();
    $available_countries = [];
    foreach ($store->get('shipping_countries') as $country_item) {
      $available_countries[] = $country_item->value;
    }

    // Pickup custom: select inline form.
    if (!self::isPickupSelected($pane_form, $form_state, $this->order)) {
      /** @var \Drupal\commerce\Plugin\Commerce\InlineForm\EntityInlineFormInterface $inline_form */
      $inline_form = $this->inlineFormManager->createInstance('customer_profile', [
        'profile_scope' => 'shipping',
        'available_countries' => $available_countries,
        'address_book_uid' => $this->order->getCustomerId(),
        // Don't copy the profile to address book until the order is placed.
        'copy_on_save' => FALSE,
      ], $this->getShippingProfilePickup());
    }
    else {
      $shipping_method = $this->getShippingMethod($pane_form, $form_state, $this->order);

      $inline_form = $this->inlineFormManager->createInstance('pickup_profile', [
        'profile_scope' => 'shipping',
        'available_countries' => $available_countries,
        'shipping_method' => $shipping_method,
      ], $this->getShippingProfilePickup(TRUE));
    }
    $pane_form['shipping_profile'] = [
      '#parents' => array_merge($pane_form['#parents'], ['shipping_profile']),
      '#inline_form' => $inline_form,
    ];
    $pane_form['shipping_profile'] = $inline_form->buildInlineForm($pane_form['shipping_profile'], $form_state);
    // The shipping_profile should always exist in form state
    if (!$form_state->has('shipping_profile')) {
      $form_state->set('shipping_profile', $inline_form->getEntity());
    }

    // Pickup custom: place at top.
    $pane_form['shipments']['#weight'] = -999;

    $shipments = $this->order->get('shipments')->referencedEntities();

    $single_shipment = count($shipments) === 1;
    foreach ($shipments as $index => $shipment) {
      /** @var \Drupal\commerce_shipping\Entity\ShipmentInterface $shipment */
      $pane_form['shipments'][$index] = [
        '#parents' => array_merge($pane_form['#parents'], ['shipments', $index]),
        '#array_parents' => array_merge($pane_form['#parents'], ['shipments', $index]),
        '#type' => $single_shipment ? 'container' : 'fieldset',
        '#title' => $shipment->getTitle(),
      ];
      $form_display = EntityFormDisplay::collectRenderDisplay($shipment, 'checkout');
      $form_display->removeComponent('shipping_profile');
      $form_display->buildForm($shipment, $pane_form['shipments'][$index], $form_state);
      $pane_form['shipments'][$index]['#shipment'] = $shipment;

      // Pickup custom: Add ajax.
      $widget = &$pane_form['shipments'][$index]['shipping_method']['widget'][0];

      $widget['#ajax'] = [
        'callback' => [self::class, 'ajaxRefreshForm'],
        'element' => $widget['#field_parents'],
      ];
      $widget['#limit_validation_errors'] = [
        $widget['#field_parents'],
      ];
    }

    return $pane_form;
  }

  /**
   * {@inheritdoc}
   */
  public function validatePaneForm(array &$pane_form, FormStateInterface $form_state, array &$complete_form) {
    $shipping_method = NestedArray::getValue(
      $form_state->getValues(),
      array_merge($pane_form['#parents'], ['shipments', 0, 'shipping_method', 0])
    );

    $is_packeta = strpos($shipping_method, 'packeta') !== FALSE;

    $profile_data = NestedArray::getValue(
      $form_state->getValues(),
      array_merge($pane_form['#parents'], ['shipping_profile'])
    );

    $is_packeta_profile = isset($profile_data['packeta']);

    // Check if chosen shipping method and profile are matching
    if ($is_packeta !== $is_packeta_profile) {
      $form_state->setError($pane_form, $this->t('Illegal input, refresh the page before continuing.'));
    }

    parent::validatePaneForm($pane_form, $form_state, $complete_form);
  }

  /**
   * Determines if a pickup shipping method is selected.
   *
   * This function is static, to not duplicate code.
   * Also used in ProfileFieldCopyWithoutPickup.
   *
   * @param array|NULL $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   *
   * @return bool
   */
  public static function isPickupSelected(?array $form, FormStateInterface $form_state, OrderInterface $order): bool {

    if ($form != NULL) {
      $shipping_method = NestedArray::getValue(
        $form_state->getUserInput(),
        array_merge($form['#parents'], ['shipments', 0, 'shipping_method', 0])
      );
    }

    if (empty($shipping_method)) {
      return self::isShipmentPickup($order);
    }

    return strpos($shipping_method, 'packeta') !== FALSE;
  }

  /**
   * Determine whether shipment shipping method is pickup
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface|null $order
   *
   * @return bool
   */
  public static function isShipmentPickup(?OrderInterface $order): bool {
    if ($order == NULL) {
      return FALSE;
    }

    /** @var \Drupal\commerce_shipping\Entity\ShipmentInterface $shipment */
    $shipments = $order->get('shipments')->referencedEntities();
    $shipment = reset($shipments);

    return $shipment !== FALSE &&
    !empty($shipment->getShippingMethod()) &&
      $shipment->getShippingMethod()
        ->getPlugin()
        ->getPluginId() === 'packeta';
  }


  /**
   * Determines current shipping method selected
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *
   * @return \Drupal\commerce_shipping\Entity\ShippingMethodInterface|\Drupal\Core\Entity\EntityInterface|null
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getShippingMethod(array $form, FormStateInterface $form_state, OrderInterface $order) {
    $shipping_method = NestedArray::getValue(
      $form_state->getUserInput(),
      array_merge($form['#parents'], ['shipments', 0, 'shipping_method', 0])
    );

    if (empty($shipping_method)) {

      /** @var \Drupal\commerce_shipping\Entity\ShipmentInterface $shipment */
      $shipments = $order->get('shipments')->referencedEntities();
      $shipment = reset($shipments);

      $shipping_method = $shipment->getShippingMethod();

      return $shipping_method;
    }

    $arr = explode('--', $shipping_method);
    $shipping_method_id = reset($arr);

    $shipping_method = $this->entityTypeManager
      ->getStorage('commerce_shipping_method')->load($shipping_method_id);

    return $shipping_method;
  }

  /**
   * Gets the shipping profile and clears address data on switch.
   *
   * @param bool $is_current_pickup
   *   Is pickup store profile.
   *
   * @return \Drupal\profile\Entity\ProfileInterface
   *   The retrieved profile.
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function getShippingProfilePickup(bool $is_current_pickup = FALSE): ProfileInterface {
    $profile = $this->getShippingProfile();

    $this->setLastShippingProfile($profile);

    $is_last_pickup = $profile->getData('is_pickup_point', FALSE);

    // Swap pickup and generic profile on shipping method change
    if ($is_current_pickup !== $is_last_pickup) {
      $last_profile = $this->getLastShippingProfile($is_current_pickup);

      if ($last_profile != NULL) {
        return $last_profile;
      }

      $profile = $this->entityTypeManager->getStorage('profile')->create([
        'type' => $profile->bundle(),
        'uid' => 0,
      ]);

      $profile->setData('is_pickup_point', $is_current_pickup);

      $this->order->save();
    }

    return $profile;
  }

  /**
   * @param \Drupal\profile\Entity\ProfileInterface $profile
   *
   * @return void
   */
  protected function setLastShippingProfile(ProfileInterface $profile) {
    if ($profile->id() != NULL) {
      if ($profile->getData('is_pickup_point', FALSE)) {
        $this->order->setData('shipping_profile.last_packeta', $profile->id());
      } else {
        $this->order->setData('shipping_profile.last_generic', $profile->id());
      }
    }
  }

  /**
   * @param bool $isPickup
   *
   * @return \Drupal\profile\Entity\ProfileInterface|null
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getLastShippingProfile(bool $isPickup = FALSE): ?ProfileInterface {
    if ($isPickup) {
      $profile_id = $this->order->getData('shipping_profile.last_packeta');
    } else {
      $profile_id = $this->order->getData('shipping_profile.last_generic');
    }

    if ($profile_id != NULL) {
      return $this->entityTypeManager->getStorage('profile')->load($profile_id);
    }

    return NULL;
  }

}
